#include <ctype.h>
#include "libmsgs.h"
#include "config.h"
#include "logging.h"
#include "callbacks.h"

#include "kom.h"
#include "earthworm.h"
#include "q6602ew.h"
#include "q660_globals.h"

enum INIT_COMMANDS
{
    INIT_MODULEID,
    INIT_RINGNAME,
    INIT_HEARTBEATINT,
    INIT_IPADDRESS,
    INIT_STATIONNAME,
    INIT_SERIALNUMBER,
    NCOMMAND
 };

const char * const INIT_COMMAND_TEXT[NCOMMAND] =
{
    "ModuleId",
    "RingName",
    "HeartbeatInt",
    "IPAddress",
    "StationName",
    "SerialNumber"
};

static int copy_str(char* destination, int destinationSize, const char* name, const char* source) {
  if (source != NULL && *source != '\0') {
    if (strlen(source) >= destinationSize) {
      logit("et", "%s name too long; max is %d\n",
        name, destinationSize - 1);
      return EW_FAILURE;
    }
    strcpy(destination, source);
  } else {
    *destination = '\0';
  }
  return 0;
}

static int get_string(char* destination, int destinationSize, const char* name) {
  return copy_str(destination, destinationSize, name, k_str());
}

/*
 * Read the config file and populate the gConfig structure
 */
int readConfig(char *configFile) {
  int i;
  int retVal = 1;
  char *str;
  char init[NCOMMAND];     // flags, one for each required command
  for ( i = 0; i < NCOMMAND; i++ )
     init[i] = 0;
  setupDefaultConfiguration();
  logInitialize(configFile, gConfig.LogFile);
  if (copy_str(gConfig.ConfigFileName, sizeof(gConfig.ConfigFileName), "configFile", configFile)) {
    return EW_FAILURE;
  }
  if(!k_open(configFile)) {
    logit("et", "Unable to open config file: %s\n", configFile);
    return -1;
  }
  while(k_rd()) {
    str = k_str();
    // ignore comment/blank lines
    if(!str || *str == '#')
      continue;
    if(k_its("ModuleId")) {
      if (get_string(gConfig.ModuleId, sizeof(gConfig.ModuleId), "ModuleId")) {
        return EW_FAILURE;
      }
      if(GetModId(gConfig.ModuleId, &QModuleId) == -1) {
        logit("et", "%s: Invalid ModuleId <%s>. \n", Q6602EW_NAME, gConfig.ModuleId );
        logit("et", "%s: Please Register ModuleId <%s> in earthworm.d!\n", Q6602EW_NAME, gConfig.ModuleId );
        return -1;
      }
      init[INIT_MODULEID] = 1;
    } else if(k_its("RingName")) {
      if (get_string(gConfig.RingName, sizeof(gConfig.RingName), "RingName")) {
        return EW_FAILURE;
      }
      gConfig.RingKey = GetKey(gConfig.RingName);
      if (gConfig.RingKey == -1) {
        return EW_FAILURE;
      }
      gConfig.send_trace = TRUE;
      init[INIT_RINGNAME] |= 1;
    } else if(k_its("MseedRingName")) {
      if (get_string(gConfig.MseedRingName, sizeof(gConfig.MseedRingName), "MseedRingName")) {
        return EW_FAILURE;
      }
      gConfig.MseedRingKey = GetKey(gConfig.MseedRingName);
      if (gConfig.MseedRingKey == -1) {
        return EW_FAILURE;
      }
      gConfig.send_mini = TRUE;
      init[INIT_RINGNAME] |= 2;
    } else if(k_its("ContinuityFileDirectory")) {
      if (get_string(gConfig.ContFileDir, sizeof(gConfig.ContFileDir) - 1, "ContinuityFileDirectory")) {
        return EW_FAILURE;
      }
      i = strlen(gConfig.ContFileDir);
      if (i > 0 && gConfig.ContFileDir[i - 1] != DIR_SLASH)
      {
        gConfig.ContFileDir[i] = DIR_SLASH;
        gConfig.ContFileDir[++i] = '\0';
      }
    } else if(k_its("HeartbeatInt")) {
      gConfig.HeartbeatInt = k_int();
      init[INIT_HEARTBEATINT] = 1;
    } else if(k_its("LogFile")) {
      gConfig.LogFile = k_int();
      logInitialize(configFile, gConfig.LogFile);
    } else if(k_its("IPAddress")) {
      if (get_string(gConfig.par_register.q660id_address, sizeof(gConfig.par_register.q660id_address), "IPAddress")) {
        return EW_FAILURE;
      }
      init[INIT_IPADDRESS] = 1;
    } else if(k_its("BasePort")) {
      gConfig.par_register.q660id_baseport = k_int();
    } else if(k_its("StationName")) {
      if (get_string(gConfig.par_create.q660id_station, sizeof(gConfig.par_create.q660id_station), "StationName")) {
        return EW_FAILURE;
      }
      init[INIT_STATIONNAME] = 1;
    } else if(k_its("SerialNumber")) {
      str = k_str();
      if(!str || *str == '\0')
        continue;
      stringtot64(str, &gConfig.par_create.q660id_serial);
      init[INIT_SERIALNUMBER] = 1;
    } else if(k_its("Password")) {
      if (get_string(gConfig.par_register.q660id_pass, sizeof(gConfig.par_register.q660id_pass), "Password")) {
        return EW_FAILURE;
      }
    } else if(k_its("StatusInterval")) {
      gConfig.statusinterval = k_int();
    } else if(k_its("LowLatPort")) {
        printf ("LowLatPort\n");
        gConfig.par_register.low_lat = TRUE;
        strcpy (gConfig.par_register.opt_lookback, "0.3");
        gConfig.dataserv.ds_port = k_int();
        gConfig.dataserv.server_number = 1;
        gConfig.dataserv.record_count = MAX_DS_BUFFERS;
        gConfig.dataserv.dsbuf = malloc (sizeof(tdsbuf));
        gConfig.dataserv.stnctx = context;
        lldatacontext = lib_ds_start (&gConfig.dataserv);
    } else if(k_its("LogLevel")) {
      // Handle the loglevel stuff here
      gConfig.LogLevel = 0;
      while ((str = k_str()) != NULL && *str != '#') {
        if (*str == ',')
          str++;
        if(!strncasecmp("SD", str, 2)) {
          gConfig.LogLevel |= VERB_SDUMP;
        } else if(!strncasecmp("RM", str, 2)) {
          gConfig.LogLevel |= VERB_REGMSG;
        } else if(!strncasecmp("VB", str, 2)) {
          gConfig.LogLevel |= VERB_LOGEXTRA;
        } else if(!strncasecmp("SM", str, 2)) {
          gConfig.LogLevel |= VERB_AUXMSG;
        } else if(!strncasecmp("PD", str, 2)) {
          gConfig.LogLevel |= VERB_PACKET;
        } else {
          logit("et", "Unexpected LogLevel value (%s)\n", str);
        }
      }
    } else {
      logit("et", "%s: Unknown config command (%s)\n", Q6602EW_NAME, k_get());
      retVal = 0;
    }
  }
  for ( i = 0; i < NCOMMAND; i++ )
  {
    if ( !init[i] )  // if missing
    {
      logit("et","%s: ERROR, no ", Q6602EW_NAME );
      for ( i = 0; i < NCOMMAND; i++ )
      {
        if ( !init[i]  ) logit("et", "%s ", INIT_COMMAND_TEXT[i] );
      }
      logit("et", "command(s) in <%s>.\n", configFile );
      return EW_FAILURE;
    }
  }
  return retVal;
}

/**
 * Set all of the config items to rational defaults
 */
void setupDefaultConfiguration() {
  memset (&gConfig, 0, sizeof(Q660Configuration)) ;
  gConfig.LogFile = 1;
  gConfig.fowner.call_fileacc = q660_file_callback ;
  gConfig.fowner.station_ptr = NULL ; /* Not required for only one station */
  gConfig.par_create.q660id_priority = 1 ;
  gConfig.par_create.opt_verbose = VERB_REGMSG ;
  gConfig.statusinterval = 180;
  gConfig.par_create.opt_client_msgs = 200 ;
  gConfig.par_create.opt_minifilter = OMF_ALL ;
  gConfig.par_create.opt_aminifilter = OMF_ALL ;
  gConfig.par_create.amini_exponent = 12 ;
  gConfig.par_create.amini_512highest = 20 ;
  gConfig.par_create.mini_embed = 1 ;
  gConfig.par_create.mini_separate = 0 ;
  copy_str(gConfig.par_register.q660id_address, sizeof(gConfig.par_register.q660id_address),
      "q660id_address", "localhost");
  gConfig.par_register.q660id_baseport = 7330 ;
  gConfig.par_register.opt_maxsps = 1000 ;
  gConfig.par_register.opt_start = 1 ;
}

void printConfigStructToLog() {
  logMessage("+++ Current Configuration:\n");
  logMessage("--- ConfigFileName: %s\n", gConfig.ConfigFileName);
  logMessage("--- ModuleId: %s\n", gConfig.ModuleId);
  logMessage("--- RingName: %s\n", gConfig.RingName);
  logMessage("--- RingKey: %d\n", gConfig.RingKey);
  logMessage("--- MseedRingName: %s\n", gConfig.MseedRingName);
  logMessage("--- ContinuityFileDirectory: %s\n", gConfig.ContFileDir);
  logMessage("--- MseedRingKey: %d\n", gConfig.MseedRingKey);
  logMessage("--- HeartbeatInt: %d\n", gConfig.HeartbeatInt);
  logMessage("--- LogFile: %d\n", gConfig.LogFile);
  logMessage("--- IPAddress: %s\n", gConfig.par_register.q660id_address);
  logMessage("--- BasePort: %d\n", gConfig.par_register.q660id_baseport);
  logMessage("--- StationName: %s\n", gConfig.par_create.q660id_station);
  logMessage("--- LowLatPort: %d\n", gConfig.dataserv.ds_port);
  logMessage("--- LogLevel: %s%s%s%s%s\n",
       gConfig.LogLevel & VERB_SDUMP ? "SD " : "",
       gConfig.LogLevel & VERB_REGMSG ? "RM " : "",
       gConfig.LogLevel & VERB_LOGEXTRA ? "VB " : "",
       gConfig.LogLevel & VERB_AUXMSG ? "SM " : "",
       gConfig.LogLevel & VERB_PACKET ? "PD " : ""
       );
}
