/*****************************************************************/
/*                                                               */
/* Name:  "latency.c"                                            */
/* Author: Andrew Good                                           */
/*                                                               */
/* Description:                                                  */
/*                                                               */
/*  Program to read waveform messages from the WAVE_RING,        */
/*  calculate latency values and write them to a LATENCY_RING.   */
/*                                                               */
/*                                                               */
/*  Three modes of operation are implemented:                    */
/*                                                               */
/*		1) When a message is detected on the Wave Ring, calculate  */
/*       the latency and output it immediately in real time.     */
/*                                                               */
/*    2) Output in real time, but force an output sample rate    */
/*       of 1 record per second.                                 */
/*                                                               */
/*    3) Use linear interpolation to calculate latency values    */
/*       to attach to each record and archive them at an         */
/*       artificially-induced rate of 1 sample per second        */
/*       (this requires storing enough data values to do the     */
/*       interpolations, precluding real-time output.            */
/*                                                               */
/*****************************************************************/

/*
 *
 * Copyright (c) 2017 California Institute of Technology.
 *
 * All rights reserved.
 *
 * This program is distributed WITHOUT ANY WARRANTY whatsoever.
 *
 * Do not redistribute this program without written permission.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <transport.h>
#include <swap.h>
#include <time_ew.h>
#include <trace_buf.h>
#include <earthworm.h>
// #include <libmseed.h>
#include <kom.h>
#include "net.h"

#define TYPE_NOTUSED 254
#define NLOGO 1			// Only one input datatype is expected (TraceBuf2)

#define RAW_MODE     0
#define FORCED_MODE  1
#define INTERP_MODE  2
#define SNW_MODE		3

/* versioning introduced by ISTI when incorporated into Earthworm repo */
#define VERSION "1.1.0 - 2019-09-30"


int debug = 0;			// Toggle debug messages
int testmode = 0;		// Test mode on/off (config file parameter)
int renameChannel = 1;		// Apply channel renaming conventions 
int slept = 0;			// 1 if a sleep has just occurred, 0 otherwise 

int sock;			// Socket       

double **inputTime;		// Array for storing input times
double **inputLat;		// Array for storing input latencies
double *refTime;		// Time (second) to compute  output latency value for
										// (used in Forced case, corresponds to nominal sample endTime)


/////////////////////////////////////////////////////////////////////////
double *scnl_lastPacketTime;	// Corresponds to scnl_time[i].last_packet_time in sniffwave
double **feedLat;		// Array for storing feed latencies
double **packet_size;		// Packet sizes for this channel
/////////////////////////////////////////////////////////////////////////


int *nSamp;			// Count of records read in for each SCNL tag
int *sleptStatus;		// For each input, track  whether it has returned from a sleep

int nTag;			// Counter for number of times a particular tag
										// has been encountered

// #define MAX_LEN_STR_SCNL 30  // Max length for a SCNL string
// #define MAX_NUM_SCNL 2000            // Max number of channels to check


// SCNL tags that we want to process
struct TAG
{
  char sta[TRACE_STA_LEN];
  char cha[TRACE_CHAN_LEN];
  char net[TRACE_NET_LEN];
  char loc[TRACE_LOC_LEN];
  int mode;
};


// Instance of TAG structure that we will be using
struct TAG *tag;

SHM_INFO region_in;		// Shared memory information about input (Wave) ring
SHM_INFO region_out;		// Shared memory information about output (Latency) ring


// Ring messages are:  "Logo + message".  In the case of a TraceBuf2 message, the
// "message" is "header + data".

MSG_LOGO getLogo;		// Message description (discovered by analysis)
										// Note: changes getLogo[NLOGO] to getLogo, and
										// reference below to &getLogo. 

MSG_LOGO inputLogo, outputLogo;	// Message descriptions ('transport.h')

long inputKey;			// Key to the transport ring we read from 
long outputKey;			// Key to the transport ring we write to  
long inputLength;		// Size that input  message is found to be
long outputLength;		// Length of output message

char inmsg[MAX_TRACEBUF_SIZ];	// Buffer to store messages from Wave Ring
char outmsg[MAX_TRACEBUF_SIZ];	// Buffer to store messages for Latency Ring   

char inRing[1024];		// Name of input ring (e.g. "WAVE_RING")
char outRing[1024];		// Name of output ring (e.g. "LATENCY_RING")
char scnlFile[1024];		// Name of SCNL list file
char hostName[1024];		// Name of host machine 
char sockName[1024];		// Name of client machine 
int portNum;			// Port number

static unsigned char InstId;	// Local installation ID (issued)
static unsigned char ModId;	// Module ID (in "earthworm.d", MOD_LATENCY=150)

TRACE2_HEADER *trh;		// Pointer to the beginning of "inmsg"
TRACE2_HEADER *trout;		// Pointer to the beginning of "outmsg"       


// Pointer(s) to beginning of message data (array address + header size)
// within "inmsg" and "outmsg".  Short (i2) values are not expected, but allowed for.  
// Test data is all of type i4.

short *short_data_in;
short *short_data_out;
int *long_data_in;
int *long_data_out;

double d_now;			// Current system clock time
double then;			// Previous heartbeat time
double timeout;			// Timeout parameter
double interval;		// Integration interval for interpolation calculations
double sleepTime;		// Amount of time to sleep when no data is detected on a read
double increment;		// Time between output records
double samprate;		// Output sampling rate 
char LocationCode[3];		// location code for output packets

double t_offset = 0;		// Amount to offset input time by when in test mode


// Output packet header values

char chan0;
char chan1;
char v0;
char v1;
char datatype[32];
// char                 quality[32];    // Quality field is more complex; we are repurposing it here
char pad[32];
int pin;
int outNsamp;


// Return index from array scnl_time for SCNL (sta, cha, net, loc). 
// If the SCNL is not present yet, it is inserted.               
// int search_scnl_time(SCNL_LAST_SCNL_PACKET_TIME *scnl_time, int *n_scnl_time, char *sta, char *cha, char *net, char *loc);

// Check if there is a gap from the last packet and the current one in trh.
// A message is printed to stderr if a gap or an overlap occurs.
// int  check_for_gap_overlap_and_output_message(TRACE2_HEADER  *trh, SCNL_LAST_SCNL_PACKET_TIME *scnl_time, int *n_scnl_time);

// Compute all packet latency values that it is possible to compute with the data
// that has been read in so far, and output those latencies to the LATENCY_RING 
int getRawLatency ();
int getForcedLatency ();
int getInterpolatedLatency ();
int getSNWLatency ();


/************************************************************************************/
/* Main method                                                                      */
/************************************************************************************/

int
main (int argc, char **argv)
{
  unsigned char InstWildcard;	// We accept messages from any installation
  unsigned char ModWildcard;	// We accept message from any module
  unsigned char Type_TraceBuf2;	// Expected input type for this module
  unsigned char Type_Heartbeat;	// Type for Earthworm heartbeat

  MSG_LOGO hrtlogo;		// Logo of outgoing heartbeats

  FILE *fin;			// Input file pointer

  char line[1024];		// String to hold SCNL tag value
  char modeString[32];		// String to hold the operational mode value
  char *strptr;			// Utility string pointer
  char *begin, *end;		// Utility character pointers
  int i;			// Utility counter
  int rc;			// Return code
  int len;			// Length of a line from SCNL input file

  int pid;			// Process ID
  int nLogo = NLOGO;		// Number of logo patterns we are willing to match
  int iTag;			// Index of the tag we are processing (within tag list)

  int defaultMode = RAW_MODE;




  double DL, TL, ML;		// Transport latency, Total latency, Mean latency
  double FL;			// Feed latency
  double heartbeatTime;		// Time between heartbeats

  unsigned char sequence_number = 0;	// Used by tport_copyfrom

  strcpy (hostName, "");
  LocationCode[0]=0;


// Check command line arguments
  if (argc < 2)
    {
      fprintf (stderr, "Error: latency - Insufficient arguments\n");
      fprintf (stderr, "Usage: latency <configFile>\n");
      fprintf (stderr, "Version: %s\n", VERSION);
      fflush (stderr);
      return -1;
    }
  else if (argc > 2)
    {
      if (strcmp (argv[2], "-d") == 0 || strcmp (argv[2], "-d1") == 0)
	{
	  debug = 1;
	}
      else if (strcmp (argv[2], "-d2") == 0)
	{
	  debug = 2;
	}
      else if (strcmp (argv[2], "-d3") == 0)
	{
	  debug = 3;
	}
      else
	{
	  fprintf (stderr, "Error: Invalid arguments.\n");
	  fflush (stderr);
	  return -1;
	}

      if (debug)
	{
	  fprintf (stderr, "Debug Level = %d.\n", debug);
	  fflush (stderr);
	}
    }

  pid = getpid ();		// Process ID does not change

  if (debug > 1)
    {
      fprintf (stderr, "Process ID = %d\n", pid);
      fflush (stderr);
    }


// Initialize pointers
  trh = (TRACE2_HEADER *) inmsg;
  trout = (TRACE2_HEADER *) outmsg;
  long_data_in = (int *) (inmsg + sizeof (TRACE2_HEADER));
  short_data_in = (short *) (inmsg + sizeof (TRACE2_HEADER));
  long_data_out = (int *) (outmsg + sizeof (TRACE2_HEADER));
  short_data_out = (short *) (outmsg + sizeof (TRACE2_HEADER));

  if (debug > 1)
    {
      fprintf (stderr, "Initialized pointers\n");
      fflush (stderr);
    }


// Read configuration file

  if (k_open (argv[1]) == 0)
    {
      fprintf (stderr, "Error: Unable to open configuration file\n");
      fflush (stderr);
      return -1;
    }

  strcpy (inRing, "");		// Initialize inRing to empty
  strcpy (outRing, "");		// Initialize outRing to empty
  strcpy (scnlFile, "");	// Initialize scnlFile to empty

  // TODO: Initialize other values to something so you can do error checks if the config file 
  // didn't have them specified.

  while (k_rd ())
    {
      strptr = k_str ();	// Get the first token from line 

      // Ignore blank lines
      if (!strptr)
	continue;

      // Ignore comments
      if (strptr[0] == '#')
	continue;

      // Copy name of Wave Ring
      if (k_its ("WaveRingName"))
	{
	  strcpy (inRing, k_str ());

	  if (debug > 1)
	    {
	      fprintf (stderr, "Wave Ring Name: %s\n", inRing);
	      fflush (stderr);
	    }
	}

      // Copy name of Latency Ring
      if (k_its ("LatencyRingName"))
	{
	  strcpy (outRing, k_str ());

	  if (debug > 1)
	    {
	      fprintf (stderr, "Latency Ring Name: %s\n", outRing);
	      fflush (stderr);
	    }
	}

      // Copy name of SCNL list file
      if (k_its ("SCNLfile"))
	{
	  strcpy (scnlFile, k_str ());

	  if (debug > 1)
	    {
	      fprintf (stderr, "SNCL tag list file: %s\n", scnlFile);
	      fflush (stderr);
	    }
	}

      // Get Module ID
      if (k_its ("ModuleId"))
	{
	  if ((strptr = k_str ()))
	    {
	      // If GetModId cannot find module name in config file, then &ModId will
	      // become set to -1, and an error will be registered.

	      if (GetModId (strptr, &ModId) == -1)
		{
		  fprintf (stderr, "Error: Invalid ModuleId <%s>. \n",
			   strptr);
		  fprintf (stderr,
			   "--> Please Register ModuleId <%s> in earthworm.d\n",
			   strptr);
		  fflush (stderr);
		  return -1;
		}
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "Module ID: %s\n", strptr);
	      fflush (stderr);
	    }
	}

      // Get operational mode
      if (k_its ("Mode"))
	{
	  strcpy (modeString, k_str ());

	  if (strcmp (modeString, "RAW_MODE") == 0)
	    defaultMode = RAW_MODE;
	  else if (strcmp (modeString, "FORCED_MODE") == 0)
	    defaultMode = FORCED_MODE;
	  else if (strcmp (modeString, "INTERP_MODE") == 0)
	    defaultMode = INTERP_MODE;
	  else if (strcmp (modeString, "SNW_MODE") == 0)
	    defaultMode = SNW_MODE;

	  if (debug > 1)
	    {
	      fprintf (stderr, "Operational Mode: %s\n", modeString);
	      fflush (stderr);
	    }
	}

      // Get timeout value
      if (k_its ("Timeout"))
	{
	  timeout = atof (k_str ());

	  if (timeout <= 0)
	    {
	      fprintf (stderr, "Error: Invalid timeout in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "Timeout interval: %f\n", timeout);
	      fflush (stderr);
	    }
	}

      // Get interval value
      if (k_its ("Interval"))
	{
	  interval = atof (k_str ());

	  if (interval <= 0)
	    {
	      fprintf (stderr,
		       "Error: Invalid interpolation interval in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "Interpolation interval: %f\n", interval);
	      fflush (stderr);
	    }
	}

      // Get sleepTime value
      if (k_its ("SleepTime"))
	{
	  sleepTime = atof (k_str ());

	  if (sleepTime <= 0)
	    {
	      fprintf (stderr, "Error: Invalid sleep time in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "Sleep interval: %f\n", sleepTime);
	      fflush (stderr);
	    }
	}

      // Get time increment value
      if (k_its ("Increment"))
	{
	  increment = atof (k_str ());

	  if (increment <= 0)
	    {
	      fprintf (stderr,
		       "Error: Invalid time increment value in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "Inter-sample time increment: %f\n",
		       increment);
	      fflush (stderr);
	    }
	}

      // Get heartbeat interval
      if (k_its ("HeartbeatTime"))
	{
	  heartbeatTime = atof (k_str ());

	  if (heartbeatTime <= 0)
	    {
	      fprintf (stderr,
		       "Error: Invalid heartbeat value in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "Heartbeat interval: %f\n", heartbeatTime);
	      fflush (stderr);
	    }
	}

      // Get channel rename flag
      if (k_its ("RenameChannels"))
	{
	  renameChannel = atoi (k_str ());

	  if (renameChannel < 0)
	    {
	      fprintf (stderr,
		       "Error: Invalid heartbeat value in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "RenameChannels: %d\n", renameChannel);
	      fflush (stderr);
	    }
	}

      ////////////////////////////////////////////////////////////////////////////////////
      // Get host machine name
      if (k_its ("Host"))
	{
	  strcpy (hostName, k_str ());

	  if (debug > 1)
	    {
	      fprintf (stderr, "Host Machine: %s\n", hostName);
	      fflush (stderr);
	    }
	}

      // Get client machine name
      if (k_its ("Client"))
	{
	  strcpy (sockName, k_str ());

	  if (debug > 1)
	    {
	      fprintf (stderr, "Client Machine: %s\n", sockName);
	      fflush (stderr);
	    }
	}

      // Get port number
      if (k_its ("Port"))
	{
	  portNum = atoi (k_str ());

	  if (portNum <= 0)
	    {
	      fprintf (stderr, "Error: Invalid port in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "portNum: %d\n", portNum);
	      fflush (stderr);
	    }
	}
      ////////////////////////////////////////////////////////////////////////////////////


      // Copy output header values
      char tempStr[32];

      if (k_its ("Chan_0"))
	{
	  strcpy (tempStr, k_str ());
	  chan0 = tempStr[0];
	}
      if (k_its ("Chan_1"))
	{
	  strcpy (tempStr, k_str ());
	  chan1 = tempStr[0];
	}
      if (k_its ("Version_0"))
	{
	  strcpy (tempStr, k_str ());
	  v0 = tempStr[0];
	}
      if (k_its ("Version_1"))
	{
	  strcpy (tempStr, k_str ());
	  v1 = tempStr[0];
	}
      if (k_its ("Datatype"))
	{
	  strcpy (datatype, k_str ());
	}
      if (k_its ("LocationCode"))
	{
	  strcpy (tempStr, k_str ());
          if (strlen(tempStr) > 2) 
            {
	      fprintf (stderr,
		       "Error: Invalid location code for output in config file, must be 2 chars: %s\n", tempStr);
	      fflush (stderr);
	      return -1;
    
	    }
	  strcpy (LocationCode, tempStr);
	}
      // if(k_its("Quality"))                 {       strcpy(quality,   k_str());     }
      if (k_its ("Pad"))
	{
	  strcpy (pad, k_str ());
	}

      //      fprintf(stderr, "Header info: %c %c %c %c %s %s %s \n", chan0, chan1, v0, v1, datatype, quality, pad); 

      // Get output sample rate value
      if (k_its ("Samprate"))
	{
	  samprate = atof (k_str ());

	  if (samprate <= 0)
	    {
	      fprintf (stderr,
		       "Error: Invalid sample rate for output in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "samprate: %f\n", samprate);
	      fflush (stderr);
	    }
	}

      // Get output nsamp
      if (k_its ("Nsamp"))
	{
	  outNsamp = atoi (k_str ());

	  if (outNsamp <= 0)
	    {
	      fprintf (stderr, "Error: Invalid nsamp in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "nSamp: %d\n", outNsamp);
	      fflush (stderr);
	    }
	}

      // Get output pinno
      if (k_its ("Pin"))
	{
	  pin = atoi (k_str ());

	  if (pin < 0)
	    {
	      fprintf (stderr, "Error: Invalid pinno in config file.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "pinno: %d\n", pin);
	      fflush (stderr);
	    }
	}

      // Get testmode indicator
      if (k_its ("TestMode"))
	{
	  testmode = atoi (k_str ());

	  if (testmode < 0 || testmode > 1)
	    {
	      fprintf (stderr, "Error: Invalid TestMode value.\n");
	      fflush (stderr);
	      return -1;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "TestMode: %d\n", testmode);
	      fflush (stderr);
	    }
	}

      ///////////////////////////////////////////////////////////////////
      // Maybe add an "output mode" configuration parameter to specify
      // whether this is to be used in EW mode or SNW mode?
      ///////////////////////////////////////////////////////////////////

    }				// end while


// Close config file
  k_close ();

  if (debug > 1)
    {
      fprintf (stderr, "Config file closed.\n");
      fflush (stderr);
    }


// Check that SCNL file has been obtained
  if (strlen (scnlFile) == 0)
    {
      fprintf (stderr, "Error: No SCNL input file specified.\n");
      fflush (stderr);
      return -1;
    }


// Get count of SCNL tags                       
  fin = fopen (scnlFile, "r");

  if (debug > 1)
    {
      fprintf (stderr, "Get SCNL tag count from %s\n", scnlFile);
      fflush (stderr);
    }

  if (fin == (FILE *) NULL)
    {
      fprintf (stderr, "Error: SCNL input file not found.\n");
      fflush (stderr);
      return -1;
    }

  nTag = 0;

  while (1)
    {
      if (fgets (line, 1024, fin) == NULL)
	break;

      ++nTag;
    }

  fclose (fin);


// Allocate space for various arrays
/*****************************************************************************/

  if (debug > 1)
    {
      fprintf (stderr, "Allocate space for various arrays.\n");
      fflush (stderr);
    }


// Allocate space for nSamp and sleptStatus
  nSamp = (int *) malloc (nTag * sizeof (int));
  sleptStatus = (int *) malloc (nTag * sizeof (int));


// Allocate space for, and set, current refTime
  hrtime_ew (&d_now);

  refTime = (double *) malloc (nTag * sizeof (double));
  scnl_lastPacketTime = (double *) malloc (nTag * sizeof (double));

  for (i = 0; i < nTag; ++i)
    {
      refTime[i] = (int) (d_now + 1);	// The next integer second value after d_now 
      scnl_lastPacketTime[i] = (int) (d_now + 1);	// The next integer second value after d_now 
    }


// Allocate space for pointers to all the inputTime and inputLat arrays
  inputTime = (double **) malloc (nTag * sizeof (double *));
  inputLat = (double **) malloc (nTag * sizeof (double *));
  feedLat = (double **) malloc (nTag * sizeof (double *));
  packet_size = (double **) malloc (nTag * sizeof (double *));


// Allocate space for each input array and initialize counts
  for (i = 0; i < nTag; ++i)
    {
      nSamp[i] = 0;
      sleptStatus[i] = 0;

      inputTime[i] = (double *) malloc (1024 * sizeof (double));	// 1024 is arbitrary
      inputLat[i] = (double *) malloc (1024 * sizeof (double));	// 1024 is arbitrary
      feedLat[i] = (double *) malloc (1024 * sizeof (double));	// 1024 is arbitrary
      packet_size[i] = (double *) malloc (1024 * sizeof (double));	// 1024 is arbitrary
    }


// Allocate space for tag structures
  tag = (struct TAG *) malloc (nTag * sizeof (struct TAG));

/*****************************************************************************/
// end array allocations


// Parse SCNL tag
/*****************************************************************************/

  if (debug > 1)
    {
      fprintf (stderr, "Parse SCNL tag.\n");
      fflush (stderr);
    }

  fin = fopen (scnlFile, "r");

  for (i = 0; i < nTag; ++i)
    {
      // Initialize SCNL to empty
      strcpy (tag[i].sta, "");
      strcpy (tag[i].cha, "");
      strcpy (tag[i].net, "");
      strcpy (tag[i].loc, "");
      tag[i].mode = defaultMode;

      fgets (line, 1024, fin);	// Read a line from scnlFile
      len = strlen (line);

      if (strstr (line, "RAW_MODE") != 0)
	tag[i].mode = RAW_MODE;
      else if (strstr (line, "FORCED_MODE") != 0)
	tag[i].mode = FORCED_MODE;
      else if (strstr (line, "INTERP_MODE") != 0)
	tag[i].mode = INTERP_MODE;
      else if (strstr (line, "SNW_MODE") != 0)
	tag[i].mode = SNW_MODE;

      begin = line;
      end = line;

	/********************************************************************************/
      // Procedure: Increment "end" pointer until it encounters a delimiter or
      // a break (i.e. '.' or '\n' or '\0') and explicitly set this character
      // to a null.  The substring between the "begin" and "end" pointers
      // (accessed via "begin") constitutes the value for the SCNL field
      // being read (e.g. Station).  
      //
      // Check that the field value read in does not exceed the maximum allowed size.  
      // Copy the value into the appropriate "tag" structure variable. Reset "begin" 
      // to point to the next substring (i.e. "end + 1") and increment "end" to point 
      // to the same place.  
      //
      // Check to see whether "end" is  pointing beyond the end of the line being read.  
      // If it is not, then repeat the process for the next substring; if it is, then 
      // move on to the next line in the file.  
      //
      // This procedure repeats for all four SCNL fields.
      // 
      // Due to the inconsistent nature of "location" values seen in
      // input files, we place a special check that replaces values with
      // a "-" in them with an empty string.
	/********************************************************************************/

      // Get station
      while (*end != '.' && *end != '\n' && *end != '\0')
	{
	  ++end;
	}

      *end = '\0';		// Replace delimiter with a null

      if (strlen (begin) > TRACE_STA_LEN)
	continue;

      strcpy (tag[i].sta, begin);

      begin = end + 1;
      end = begin;

      if ((end - line) >= len)
	continue;

      // Get channel
      while (*end != '.' && *end != '\n' && *end != '\0')
	{
	  ++end;
	}

      *end = '\0';		// Replace delimiter with a null

      if (strlen (begin) > TRACE_CHAN_LEN)
	continue;

      strcpy (tag[i].cha, begin);

      begin = end + 1;
      end = begin;

      if ((end - line) >= len)
	continue;

      // Get network
      while (*end != '.' && *end != '\n' && *end != '\0')
	{
	  ++end;
	}

      *end = '\0';		// Replace delimiter with a null

      if (strlen (begin) > TRACE_NET_LEN)
	continue;

      strcpy (tag[i].net, begin);

      begin = end + 1;
      end = begin;

      if ((end - line) >= len)
	continue;

      // Get location
      while (*end != '.' && *end != '\n' && *end != '\0')
	{
	  ++end;
	}

      *end = '\0';		// Replace delimiter with a null

      if (strlen (begin) > TRACE_LOC_LEN)
	continue;

      strcpy (tag[i].loc, begin);

      //      if(strstr(begin, "-") != 0)
      //              strcpy(tag[i].loc, "");
    }

/*****************************************************************************/
// end SCNL tag parsing


  fclose (fin);			// Close tag list file

  if (debug > 1)
    {
      for (i = 0; i < nTag; ++i)
	{
	  fprintf (stderr, "tag%d:[%s][%s][%s][%s][%d]\n",
		   i, tag[i].sta, tag[i].cha, tag[i].net, tag[i].loc,
		   tag[i].mode);
	  fflush (stderr);
	}
    }

// Check that Ring names have been obtained
  if (strlen (inRing) == 0)
    {
      fprintf (stderr, "Error: No input ring name found; exiting...\n");
      fflush (stderr);
      return -1;
    }
  if (strlen (outRing) == 0 && defaultMode != SNW_MODE)
    {
      fprintf (stderr, "Error: No ring names found; exiting...\n");
      fflush (stderr);
      return -1;
    }


// Attach to input ring
  if ((inputKey = GetKey (inRing)) == -1)
    {
      fprintf (stderr, "Error: Invalid RingName; exiting...\n");
      fflush (stderr);
      return -1;
    }
  tport_attach (&region_in, inputKey);

  if (debug > 1)
    {
      fprintf (stderr, "Attached to input ring.\n");
      fflush (stderr);
    }


// Attach to output ring
  if (defaultMode != SNW_MODE)
    {
      if ((outputKey = GetKey (outRing)) == -1)
	{
	  fprintf (stderr, "Error: Invalid RingName; exiting...\n");
	  fflush (stderr);
	  return -1;
	}
      tport_attach (&region_out, outputKey);

      if (debug > 1)
	{
	  fprintf (stderr, "Attached to output ring.\n");
	  fflush (stderr);
	}
    }

// Look up local installation id
  if (GetLocalInst (&InstId) != 0)
    {
      fprintf (stderr, "Error getting local installation id; exiting...\n");
      fflush (stderr);
      return -1;
    }


// Specify logos to get
  if (GetType ("TYPE_TRACEBUF2", &Type_TraceBuf2) != 0)
    {
      fprintf (stderr,
	       "%s: Invalid message type <TYPE_TRACEBUF2>! Missing from earthworm.d or earthworm_global.d\n",
	       argv[0]);
      fflush (stderr);
      return -1;
    }

  if (GetType ("TYPE_HEARTBEAT", &Type_Heartbeat) != 0)
    {
      fprintf (stderr, "%s: Error getting Type_Heartbeat.\n", argv[0]);
      fflush (stderr);
      return -1;
    }

  if (GetModId ("MOD_WILDCARD", &ModWildcard) != 0)
    {
      fprintf (stderr,
	       "%s: Invalid moduleid <MOD_WILDCARD>! Missing from earthworm.d or earthworm_global.d\n",
	       argv[0]);
      fflush (stderr);
      return -1;
    }

  if (GetInst ("INST_WILDCARD", &InstWildcard) != 0)
    {
      fprintf (stderr,
	       "%s: Invalid instid <INST_WILDCARD>! Missing from earthworm.d or earthworm_global.d\n",
	       argv[0]);
      fflush (stderr);
      return -1;
    }

  if (debug > 1)
    {
      fprintf (stderr, "Construct input and output messages.\n");
      fflush (stderr);
    }


// Construct input message logo
  getLogo.type = Type_TraceBuf2;
  getLogo.instid = InstWildcard;
  getLogo.mod = ModWildcard;


// Construct output message logo 
  if (defaultMode != SNW_MODE)
    {
      outputLogo.type = Type_TraceBuf2;
      outputLogo.mod = ModId;
      outputLogo.instid = InstId;
    }
  else
    {
      sock = client_connect (sockName, portNum);	// Change these (name and port) to config parameters

      if (sock == -1)
	{
	  fprintf (stderr, "Error: client_connect(%s, %d) return = %d\n",
		   sockName, portNum, sock);
	}

      int tbl = 1;
      if (tbl)
	{
	  if (debug)
	    {
	      printf ("sock = %d\n", sock);
	    }

	  printf
	    ("| NETSTA | P |           DATETIME   |       DL   |       FL     | T_SINCE  |    P_SIZE  |\n");
	  fflush (stdout);
	}

    }

// Flush the ring
  while (tport_copyfrom
	 (&region_in, &getLogo, nLogo, &inputLogo, &inputLength, inmsg,
	  MAX_TRACEBUF_SIZ, &sequence_number) != GET_NONE)
    {
      // sleep_ew(10);
      // fprintf(stderr, "Flushing record\n");
      // NO-OP
    }

///////////////////////////////////////////// OFFSET
// TPORT_LOCATION *oldLocation;
// TPORT_HEAD     *tmphd;
// FILE           *foff;
// long             offset;
// char            offString[1024];
///////////////////////////////////////////// OFFSET

// Main Loop 
  while (tport_getflag (&region_in) != TERMINATE)
    {
      ///////////////////////////////////////////// OFFSET
      // oldLocation = tport_getLocation();
      // tmphd = (TPORT_HEAD *)(&oldLocation->header);

      // if((foff=fopen("offLoc", "r")) != (FILE *)NULL)
      // {
      //      fgets(offString, 1024, foff);

      //      offset = atol(offString);
      //      fprintf(stderr, "Offset read: %ld\n", offset); 
      //      fflush(stderr);
      //      fclose(foff);
      //      unlink("offLoc");
      //      tport_setLocation(offset);
      // }
      // else
      // {
      //      // printf("ERROR: foff = %p\n", foff);
      // }
      ///////////////////////////////////////////// OFFSET

      rc =
	tport_copyfrom (&region_in, &getLogo, nLogo, &inputLogo, &inputLength,
			inmsg, MAX_TRACEBUF_SIZ, &sequence_number);

      // Print input header
      if (debug > 2)
	{
	  //      fprintf( stderr, "tport_copyfrom rc=%d\n", rc);
	  fprintf (stderr, "Input Record --> ");
	  fprintf (stderr, "%5s.%s.%s.%s (0x%x 0x%x) ",
		   trh->sta, trh->chan, trh->net, trh->loc, trh->version[0],
		   trh->version[1]);
	  fprintf (stderr, "%d %s %3d ", trh->pinno, trh->datatype,
		   trh->nsamp);
	  fprintf (stderr, "%6.4f", trh->samprate);
	  fprintf (stderr, " (%.4f) (%.4f) ", trh->starttime, trh->endtime);
	  fprintf (stderr, "\n");
	  fflush (stderr);
	}

      if (rc == GET_NONE)
	{
	  sleep_ew (sleepTime);
	  slept = 1;

	  for (i = 0; i < nTag; ++i)
	    {
	      sleptStatus[i] = 1;
	    }

	  if (debug > 1)
	    {
	      //      fprintf( stderr, "tport_copyfrom rc=GET_NONE\n");
	      fprintf (stderr, "Slept for %f.\n", sleepTime);
	      fflush (stderr);
	    }

	  // After the sleep we want to double check to see if something should be output
	  // in the Forced and Interpolated modes before we overwrite it with a new record

	  if (defaultMode == SNW_MODE)
	    {
	      getSNWLatency ();
	    }
	  else
	    {
	      getForcedLatency ();	// Output latency values forced to 1-sec intervals 
	      getInterpolatedLatency ();	// Calculate latency values for archiving 
	    }

	  continue;
	}

      ///////////////////////////////////////////// OFFSET
      // if(debug > 2)
      // {
      //      fprintf(stderr, "Latency: Location --> ");
      //      fprintf(stderr, "[%8ld]", oldLocation->location);
      //      fprintf(stderr, " [%3d][%5ld][%d][%d][%d][%3d]\n",
      //                      tmphd->start, tmphd->size, tmphd->logo.type,
      //                      tmphd->logo.mod, tmphd->logo.instid, tmphd->seq);
      //      fflush(stderr);
      //      /*
      //      for (i=0; i<tmphd->size; ++i) 
      //      {
      //              printf("[%u]", (unsigned char) oldLocation->message[i]);
      //      }
      //      printf("\n]\n");
      //      */
      // }
      ///////////////////////////////////////////// OFFSET

      if (rc == GET_TOOBIG)
	{
	  fprintf (stderr,
		   "Error: retrieved message too big (%ld) for inmsg\n",
		   inputLength);
	  fflush (stderr);
	  continue;
	}

      if (rc == GET_NOTRACK)
	fprintf (stderr, "Error: Tracking error.\n");

      if (rc == GET_MISS_LAPPED)
	fprintf (stderr, "Error: Got lapped on the ring.\n");

      if (rc == GET_MISS_SEQGAP)
	fprintf (stderr, "Error: Gap in sequence numbers\n");

      if (rc == GET_MISS)
	fprintf (stderr, "Error: Missed messages\n");

      fflush (stderr);


      // Find the inputTime and inputLat arrays that match current record tag

      iTag = -1;

      for (i = 0; i < nTag; ++i)
	{
	  if (strcmp (trh->sta, tag[i].sta) == 0
	      && strcmp (trh->chan, tag[i].cha) == 0
	      && strcmp (trh->net, tag[i].net) == 0
	      && ((strcmp (trh->loc, tag[i].loc) == 0)
		  || (strcmp (tag[i].loc, "") == 0)))
	    {
	      iTag = i;		// If match found for SCNL tag, set value and break
	      break;
	    }
	}

      if (iTag == -1)
	continue;		// If no input is found for this tag at this time,
      // continue to next loop iteration


      //      Get current time
      hrtime_ew (&d_now);


      // Testing and debugging clause: set t_offset
      if (testmode && t_offset == 0)
	{
	  t_offset = d_now - trh->endtime;

	  if (debug > 1)
	    {
	      fprintf (stderr, "t_offset: %f\n", t_offset);
	    }
	}


      // Compute latencies
      DL = d_now - (trh->endtime + t_offset);
      TL = d_now - (trh->starttime + t_offset);
      ML = (DL + TL) / 2.;
      FL = d_now - scnl_lastPacketTime[iTag];


      // Add read values to our arrays
      // >>> arrayName[index of current tag][current count associated with that tag]
      inputTime[iTag][nSamp[iTag]] = trh->endtime;
      inputLat[iTag][nSamp[iTag]] = DL;
      scnl_lastPacketTime[iTag] = d_now;
      packet_size[iTag][nSamp[iTag]] = (trh->endtime - trh->starttime);
      feedLat[iTag][nSamp[iTag]] = FL;

      if (debug > 1)
	{
	  fprintf (stderr, "\nMain loop: iTag=%d, nSamp[iTag]=%d   -->  ",
		   iTag, nSamp[iTag]);
	  //      fprintf(stderr, "trh->endtime: %f\n", trh->endtime);
	  fprintf (stderr, "DL=%f, FL=%f\n", DL, FL);
	  fflush (stderr);
	}

      ++nSamp[iTag];		// Increment count associated with that particular tag only

      if (tag[i].mode == RAW_MODE)
	getRawLatency ();	// Output real-time latency values 


      // Process these regardless (to see whether any of the tags
      // marked with these modes are in the middle of a gap)

      if (defaultMode == SNW_MODE)
	{
	  getSNWLatency ();
	}
      else
	{
	  getForcedLatency ();	// Output latency values forced to 1-sec intervals 
	  getInterpolatedLatency ();	// Calculate latency values for archiving 
	}


      // Send a heartbeat to the output ring 
      if (defaultMode != SNW_MODE)
	{
	  if ((d_now - then) >= heartbeatTime)
	    {
	      int lineLen;
	      char line[40];

	      then = d_now;

	      sprintf (line, "%ld %d\n", (long) d_now, pid);
	      lineLen = strlen (line);

	      if (tport_putmsg (&region_out, &hrtlogo, lineLen, line) !=
		  PUT_OK)
		{
		  fprintf (stderr, "Error sending heartbeat. Exiting.\n");
		  break;
		}
	    }
	}
    }				// end while

  return 0;
}				// end main

/************************************************************************************/


/************************************************************************************/
/* Function: getRawLatency                                                          */
/************************************************************************************/

int
getRawLatency ()
{
  int j;			// Utility counter

  if (debug > 2)
    {
      fprintf (stderr, "getRawLatency()\n");
      fflush (stderr);
    }

  for (j = 0; j < nTag; ++j)	// 'j' is the index of the tag we are considering
    {
      if (nSamp[j] > 0 && tag[j].mode == RAW_MODE)
	{

	  // Convert output latency to integer rounded to the nearest millisecond
	  *long_data_out = (int) (1000 * inputLat[j][0] + 0.5);


	  // Populate trout and apply naming convention rules
	  strcpy (trout->sta, tag[j].sta);
	  strcpy (trout->chan, tag[j].cha);
	  strcpy (trout->net, tag[j].net);
	  strcpy (trout->loc, tag[j].loc);
          if (LocationCode[0] != 0)
            {
	       strcpy (trout->loc, LocationCode);
            }
	  if (renameChannel)
	    {
	      trout->chan[0] = chan0;
	      if (trout->chan[1] == 'H')
		{
		  trout->chan[1] = chan1;
		}
	      if (trout->chan[1] == 'N')
		{
		  trout->chan[1] = chan1;
		}
	    }

	  trout->version[0] = v0;
	  trout->version[1] = v1;
	  trout->pinno = pin;
	  trout->nsamp = outNsamp;
	  trout->samprate = samprate;
	  trout->endtime = inputTime[j][0] + t_offset;	// t_offset makes it relative to d_now; delete later
	  trout->starttime = trout->endtime - increment;
	  strcpy (trout->datatype, datatype);
	  // strcpy(trout->quality, quality);
	  strcpy (trout->pad, pad);

	  outputLength = sizeof (TRACE2_HEADER) + 4 * trout->nsamp;	// floating-point numbers  are 4 bytes eac?


	  // Print output header
	  if (debug)
	    {
	      fprintf (stderr, "Raw Latency --> ");
	      fprintf (stderr, "%5s.%s.%s.%s (0x%x 0x%x) ",
		       trout->sta, trout->chan, trout->net, trout->loc,
		       trout->version[0], trout->version[1]);
	      fprintf (stderr, "%d %s %3d ", trout->pinno, trout->datatype,
		       trout->nsamp);
	      fprintf (stderr, "%6.4f", trout->samprate);
	      fprintf (stderr, " (%.4f) (%.4f) ", trout->starttime,
		       trout->endtime);
	      fprintf (stderr, "%d", long_data_out[0]);
	      fprintf (stderr, " [DL:%fs, FL:%fs] ", inputLat[j][0],
		       feedLat[j][0]);

	      if (debug > 1)
		{
		  fprintf (stderr, " --> sleptStatus[%d]=%d ", j,
			   sleptStatus[j]);
		}

	      if (slept == 1)
		{
		  fprintf (stderr, " Slept for %f ", sleepTime);
		}

	      fprintf (stderr, "\n");
	      fflush (stderr);
	    }

	  if (slept == 1)
	    slept = 0;

	  trout->quality[0] = 0;
	  trout->quality[1] = 0;

	  if (sleptStatus[j] == 1)
	    trout->quality[1] = 1;

	  // Write output message to latency ring
	  tport_putmsg (&region_out, &outputLogo, outputLength, outmsg);

	  sleptStatus[j] = 0;
	  nSamp[j] = 0;

	}			// end if
    }				// end for (each tag)
  return 0;
}				// end getRawLatency()

/************************************************************************************/


/************************************************************************************/
/* Function: getForcedLatency                                                       */
/************************************************************************************/

int
getForcedLatency ()
{
  int j;			// Utility counter

  if (debug > 2)
    {
      fprintf (stderr, "getForcedLatency(), ");
      fprintf (stderr, "  d_now: %f\n", d_now);
      fflush (stderr);
    }

  hrtime_ew (&d_now);

  for (j = 0; j < nTag; ++j)	// 'j' is the index of the tag we are considering
    {
      if (nSamp[j] == 0)
	{
	  continue;
	}

      if (d_now - (inputTime[j][nSamp[j] - 1] + t_offset) > timeout)
	{

	  if (debug > 2)
	    {
	      fprintf (stderr, "  d_now: %f\n", d_now);
	      fprintf (stderr, "  t_offset: %f\n", t_offset);
	      fprintf (stderr, "  inputTime[j][nSamp[j]-1]: %f\n",
		       inputTime[j][nSamp[j] - 1]);
	    }

	  nSamp[j] = 0;

	  if (debug > 1)
	    {
	      fprintf (stderr, "Timeout reached; continuing...\n");
	      fflush (stderr);
	    }

	  continue;
	}

      if (nSamp[j] > 0 && tag[j].mode == FORCED_MODE)
	{
	  if (refTime[j] > d_now)
	    {
	      if (debug > 1)
		{
		  fprintf (stderr, "refTime[%d]=%f, d_now=%f\n", j,
			   refTime[j], d_now);
		}

	      break;
	    }

	  // Convert output latency to integer rounded to the nearest millisecond
	  *long_data_out = (int) (1000 * inputLat[j][nSamp[j] - 1] + 0.5);


	  // Populate trout and apply naming convention rules
	  strcpy (trout->sta, tag[j].sta);
	  strcpy (trout->chan, tag[j].cha);
	  strcpy (trout->net, tag[j].net);
	  strcpy (trout->loc, tag[j].loc);
          if (LocationCode[0] != 0)
            {
	       strcpy (trout->loc, LocationCode);
            }

	  if (renameChannel)
	    {
	      trout->chan[0] = chan0;
	      if (trout->chan[1] == 'H')
		{
		  trout->chan[1] = chan1;
		}
	      if (trout->chan[1] == 'N')
		{
		  trout->chan[1] = chan1;
		}
	    }

	  trout->version[0] = v0;
	  trout->version[1] = v1;
	  trout->pinno = pin;
	  trout->nsamp = outNsamp;
	  trout->samprate = samprate;
	  trout->endtime = refTime[j];
	  trout->starttime = trout->endtime - increment;
	  strcpy (trout->datatype, datatype);
	  // strcpy(trout->quality, quality);
	  strcpy (trout->pad, pad);

	  outputLength = sizeof (TRACE2_HEADER) + 4 * trout->nsamp;


	  // Print output header
	  if (debug)
	    {
	      fprintf (stderr, "Forced Latency --> ");

	      fprintf (stderr, "%5s.%s.%s.%s (0x%x 0x%x) ",
		       trout->sta, trout->chan, trout->net, trout->loc,
		       trout->version[0], trout->version[1]);
	      fprintf (stderr, "%d %s %3d ", trout->pinno, trout->datatype,
		       trout->nsamp);
	      fprintf (stderr, "%6.4f", trout->samprate);
	      fprintf (stderr, " (%.4f) (%.4f) ", trout->starttime,
		       trout->endtime);
	      fprintf (stderr, "%8d", long_data_out[0]);
	      fprintf (stderr, " [DL:%fs, FL:%fs] ",
		       inputLat[j][nSamp[j] - 1], feedLat[j][nSamp[j] - 1]);

	      if (slept == 1)
		{
		  fprintf (stderr, " Slept for %f ", sleepTime);
		}

	      fprintf (stderr, "\n");
	      fflush (stderr);
	    }

	  if (slept == 1)
	    slept = 0;

	  trout->quality[0] = 0;
	  trout->quality[1] = 0;

	  if (sleptStatus[j] == 1)
	    trout->quality[1] = 1;

	  //      if(sleptStatus[j] == 1)
	  //              strcpy(trout->quality, "1");

	  // Write output message to latency ring
	  tport_putmsg (&region_out, &outputLogo, outputLength, outmsg);

	  sleptStatus[j] = 0;

	  inputTime[j][0] = inputTime[j][nSamp[j] - 1];
	  inputLat[j][0] = inputLat[j][nSamp[j] - 1];

	  nSamp[j] = 1;

	  // Increment the reference time
	  refTime[j] += increment;

	  // If we do not have data to fill in intermediate times, 
	  // shift the reference time that we are looking for.
	  while (refTime[j] < (inputTime[j][0] + t_offset))
	    {
	      trout->endtime = refTime[j];
	      trout->starttime = trout->endtime - increment;

	      // Write output message to latency ring
	      tport_putmsg (&region_out, &outputLogo, outputLength, outmsg);

	      refTime[j] += increment;
	    }

	  if (debug > 1)
	    {
	      fprintf (stderr, "refTime[%d]=%f\n", j, refTime[j]);
	    }

	}			// end if
    }				// end for (each tag)
  return 0;
}				// end getForcedLatency()

/************************************************************************************/


/************************************************************************************/
/* Function: getInterpolatedLatency                                                 */
/************************************************************************************/

int
getInterpolatedLatency ()
{
  int nShift;			// Count used to discard already-processed inputs
  int iBefore, iAfter;		// Indices of inputs bracketing interval of interest
  int nOut;			// Utility counter
  int i, j;			// Utility counter
  double tStart, tEnd;		// Start/end times for the interval being integrated over
  double time[1024];		// Times for point being integrated over
  double latency[1024];		// Latencies for point being integrated over
  double outputLatency;		// Latency value calculated for a given refTime
  double frac;			// Fractional value used interpolation calculation

  if (debug > 2)
    {
      fprintf (stderr, "getInterpolatedLatency()\n");
      fflush (stderr);
    }

  hrtime_ew (&d_now);

  for (j = 0; j < nTag; ++j)	// 'j' is the index of the tag we are considering
    {
      if (tag[j].mode != INTERP_MODE)
	continue;

      if (nSamp[j] == 0)
	continue;

      if (d_now - (inputTime[j][nSamp[j] - 1] + t_offset) > timeout)
	{
	  nSamp[j] = 0;
	  continue;
	}

      while (1)
	{
	  // Set interval start and end for the iteration based on refTime
	  tStart = refTime[j] - 0.5 * interval;
	  tEnd = refTime[j] + 0.5 * interval;

	  if (tStart < (inputTime[j][0] + t_offset))
	    {
	      refTime[j] += interval;

	      if (debug > 1)
		{
		  fprintf (stderr,
			   "First point, invalid tStart; continuing...\n");
		}

	      continue;
	    }

	  if (debug > 1)
	    {
	      printf ("refTime[%d]: %11.2f\n", j, refTime[j]);
	      printf ("tStart: %11.2f\n", tStart);
	      printf ("tEnd:   %11.2f\n", tEnd);
	    }


	  // Analyze arrays to determine whether we can output anything

	  iBefore = 0;
	  iAfter = 0;

	  for (i = 0; i < nSamp[j]; ++i)	// For the records currently in our sample window
	    {
	      if ((inputTime[j][i] + t_offset) <= tStart)
		{
		  // gotBefore = 1;

		  iBefore = i;	// Index of the last data value before (or at) tStart

		  if (debug > 2)
		    fprintf (stderr,
			     "inputTime[%d][%d] + t_offset <= tStart; iBefore=%d\n",
			     j, i, i);
		}
	    }


	  // Shift values in the arrays (discard values that are not necessary
	  // for future processing, to save space).  Any future values of 'tStart'
	  // will be later than the current value of 'tStart', which is itself later
	  // than the time pointed to by 'iBefore'.  Records preceding 'iBefore' 
	  // are therefore unnecessary.

	  nShift = iBefore;

	  for (i = nShift; i < nSamp[j]; ++i)
	    {
	      inputTime[j][i - nShift] = inputTime[j][i];
	      inputLat[j][i - nShift] = inputLat[j][i];
	    }

	  nSamp[j] -= nShift;	// Shift sampling window
	  iBefore = 0;		// Reset index once shift has been performed

	  for (i = 0; i < nSamp[j]; ++i)	// For the records currently in our sample window
	    {
	      if ((inputTime[j][i] + t_offset) >= tEnd)
		{
		  iAfter = i;	// Index of the first data value after (or at) tEnd

		  if (debug > 2)
		    fprintf (stderr,
			     "inputTime[%d][%d] + t_offset >= tEnd; iAfter=%d; breaking\n",
			     j, i, i);

		  break;
		}
	    }

	  if (iBefore == iAfter && iAfter < (nSamp[j] - 1))
	    ++iAfter;

	  if (debug > 1)
	    {
	      printf ("Held Input Times: ");

	      for (i = 0; i < nSamp[j]; ++i)
		printf ("%11.2f ", (inputTime[j][i] + t_offset));

	      printf ("\n");

	      printf ("iBefore=%d, iAfter=%d\n", iBefore, iAfter);
	      fflush (stdout);
	    }

	  if ((inputTime[j][iAfter] + t_offset) < tEnd)	// If pointer for iAfter is not beyond tEnd,
	    break;		// then we do not have enough data to do the
	  // next interpolation, so we need to break 
	  // out of this function and wait for another
	  // read.


	  // Calculate interpolation weighting fraction for beginning of time window
	  frac =
	    (tStart -
	     (inputTime[j][iBefore] + t_offset)) / (inputTime[j][iBefore +
								 1] -
						    inputTime[j][iBefore]);

	  if (debug > 1)
	    {
	      printf ("frac(begin)=%f\n", frac);
	    }

	  if (frac < 0 || frac > 1.0)
	    {
	      printf ("Error: Invalid fraction ( tStart ).\n");
	      exit (0);
	    }

	  nOut = 0;		// Reset counter at beginning of time window


	  // Apply weighting and add interpolated latency to array used for calculations
	  time[nOut] = tStart;
	  latency[nOut] =
	    frac * inputLat[j][iBefore + 1] + (1. -
					       frac) * inputLat[j][iBefore];
	  ++nOut;


	  // Add points entirely within time window to the arrays used for calculations
	  for (i = iBefore + 1; i < iAfter; ++i)
	    {
	      time[nOut] = inputTime[j][i] + t_offset;
	      latency[nOut] = inputLat[j][i];
	      ++nOut;
	    }


	  // Re-calculate interpolation weighting fraction for end of time window
	  frac =
	    (tEnd -
	     (inputTime[j][iAfter - 1] + t_offset)) / (inputTime[j][iAfter] -
						       inputTime[j][iAfter -
								    1]);

	  if (debug > 1)
	    {
	      printf ("frac(end)=%f\n", frac);
	    }

	  if (frac < 0 || frac > 1.0)
	    {
	      printf ("Error: Invalid fraction ( tEnd ).\n");
	      exit (0);
	    }


	  // Apply weighting and add interpolated latency to array used for calculations
	  time[nOut] = tEnd;
	  latency[nOut] =
	    frac * inputLat[j][iAfter] + (1. - frac) * inputLat[j][iAfter -
								   1];
	  ++nOut;

	  if (debug > 1)
	    {
	      for (i = 0; i < nOut; ++i)
		{
		  printf ("time[%d]=%f, latency[%d]=%f\n", i, time[i], i,
			  latency[i]);
		}
	    }


	  // Calculate mean latency as trapezoidal summation over current time window
	  outputLatency = 0;

	  for (i = 1; i < nOut; ++i)
	    {
	      outputLatency +=
		0.5 * (latency[i] + latency[i - 1]) * (time[i] - time[i - 1]);
	    }

	  outputLatency /= interval;

	  if (debug > 1)
	    {
	      printf
		("Tag=%d, iBefore=%d, iAfter=%d: refTime=%11.2f, outputLatency= %7.4f, nshift=%d\n",
		 j, iBefore, iAfter, refTime[j], outputLatency, iBefore + 1);
	    }

	  if (debug > 1)
	    {
	      printf ("Time: %11.2f, Latency: %7.4f\n", refTime[j],
		      outputLatency);
	      fflush (stdout);
	    }


	  // Convert output latency to integer rounded to the nearest millisecond
	  *long_data_out = (int) (1000 * outputLatency + 0.5);


	  // Populate trout and apply naming convention rules
	  strcpy (trout->sta, tag[j].sta);
	  strcpy (trout->chan, tag[j].cha);
	  strcpy (trout->net, tag[j].net);
	  strcpy (trout->loc, tag[j].loc);
          if (LocationCode[0] != 0)
            {
	       strcpy (trout->loc, LocationCode);
            }

	  if (renameChannel)
	    {
	      trout->chan[0] = chan0;
	      if (trout->chan[1] == 'H')
		{
		  trout->chan[1] = chan1;
		}
	      if (trout->chan[1] == 'N')
		{
		  trout->chan[1] = chan1;
		}
	    }

	  trout->version[0] = v0;
	  trout->version[1] = v1;
	  trout->pinno = pin;
	  trout->nsamp = outNsamp;
	  trout->samprate = samprate;
	  trout->endtime = refTime[j];
	  trout->starttime = trout->endtime - increment;
	  strcpy (trout->datatype, datatype);
	  // strcpy(trout->quality, quality);
	  strcpy (trout->pad, pad);

	  outputLength = sizeof (TRACE2_HEADER) + 4 * trout->nsamp;


	  // Print output header
	  if (debug)
	    {
	      fprintf (stderr, "Interpolated Latency --> ");
	      fprintf (stderr, "%5s.%s.%s.%s (0x%x 0x%x) ",
		       trout->sta, trout->chan, trout->net, trout->loc,
		       trout->version[0], trout->version[1]);
	      fprintf (stderr, "%d %s %3d ", trout->pinno, trout->datatype,
		       trout->nsamp);
	      fprintf (stderr, "%6.4f", trout->samprate);
	      fprintf (stderr, " (%.4f) (%.4f) ", trout->starttime,
		       trout->endtime);
	      fprintf (stderr, "%d", long_data_out[0]);

	      if (slept == 1)
		{
		  fprintf (stderr, " Slept for %f ", sleepTime);
		}

	      fprintf (stderr, "\n");
	      fflush (stderr);
	    }

	  if (slept == 1)
	    slept = 0;

	  trout->quality[0] = 0;
	  trout->quality[1] = 0;

	  if (sleptStatus[j] == 1)
	    trout->quality[1] = 1;

	  //      if(sleptStatus[j] == 1)
	  //              strcpy(trout->quality, "1");

	  // Write output message to latency ring
	  tport_putmsg (&region_out, &outputLogo, outputLength, outmsg);

	  sleptStatus[j] = 0;


	  // Increment the reference time
	  refTime[j] += increment;


	  // If we do not have data to fill in intermediate times, 
	  // shift the reference time that we are looking for.
	  if (refTime[j] < (inputTime[j][0] + t_offset))
	    {
	      refTime[j] = (int) ((inputTime[j][0] + t_offset) + increment);
	    }
	}			// end while (processing a time window)
    }				// end for (each tag)
  return 0;
}				// end getInterpolatedLatency

/************************************************************************************/



/************************************************************************************/
/* Function: getSNWLatency                                                          */
/************************************************************************************/

int
getSNWLatency ()
{
  int nShift;			// Count used to discard already-processed inputs
  int iBefore, iAfter;		// Indices of inputs bracketing interval of interest
  int nOut;			// Utility counter
  int i, j;			// Utility counter
  double tStart, tEnd;		// Start/end times for the interval being integrated over
  double time[1024];		// Times for point being integrated over
  double dlatency[1024];	// Data latencies (DL) for point being integrated over
  double flatency[1024];	// Feed latencies (FL) for point being integrated over
  double psize[1024];		// Packet sizes for point being integrated over
  double outputDLatency;	// Data latency value calculated for a given refTime
  double outputFLatency;	// Feed latency value calculated for a given refTime
  double avPacketSize;		// Average packet size
  double frac;			// Fractional value used interpolation calculation
  double time_since_last = 0.0;
  double flatmin = 0.01;

  if (debug > 2)
    {
      fprintf (stderr, "getSNWLatency()\n");
      fflush (stderr);
    }


  hrtime_ew (&d_now);

  for (j = 0; j < nTag; ++j)	// 'j' is the index of the tag we are considering
    {
      if (tag[j].mode != SNW_MODE)
	continue;

      if (nSamp[j] == 0)
	continue;

      if (d_now - (inputTime[j][nSamp[j] - 1] + t_offset) > timeout)
	{
	  nSamp[j] = 0;
	  continue;
	}

      while (1)
	{
	  // Set interval start and end for the iteration based on refTime
	  tStart = refTime[j] - 0.5 * interval;
	  tEnd = refTime[j] + 0.5 * interval;

	  if (tStart < (inputTime[j][0] + t_offset))
	    {
	      refTime[j] += interval;

	      if (debug > 1)
		{
		  fprintf (stderr,
			   "First point, invalid tStart; continuing...\n");
		}

	      continue;
	    }

	  if (debug > 1)
	    {
	      printf ("refTime[%d]: %11.2f\n", j, refTime[j]);
	      printf ("tStart: %11.2f\n", tStart);
	      printf ("tEnd:   %11.2f\n", tEnd);
	    }


	  // Analyze arrays to determine whether we can output anything

	  iBefore = 0;
	  iAfter = 0;

	  for (i = 0; i < nSamp[j]; ++i)	// For the records currently in our sample window
	    {
	      if ((inputTime[j][i] + t_offset) <= tStart)
		{
		  // gotBefore = 1;

		  iBefore = i;	// Index of the last data value before (or at) tStart

		  if (debug > 2)
		    fprintf (stderr,
			     "inputTime[%d][%d] + t_offset <= tStart; iBefore=%d\n",
			     j, i, i);
		}
	    }


	  // Shift values in the arrays (discard values that are not necessary
	  // for future processing, to save space).  Any future values of 'tStart'
	  // will be later than the current value of 'tStart', which is itself later
	  // than the time pointed to by 'iBefore'.  Records preceding 'iBefore' 
	  // are therefore unnecessary.

	  nShift = iBefore;

	  for (i = nShift; i < nSamp[j]; ++i)
	    {
	      inputTime[j][i - nShift] = inputTime[j][i];
	      inputLat[j][i - nShift] = inputLat[j][i];
	      feedLat[j][i - nShift] = feedLat[j][i];
	    }

	  nSamp[j] -= nShift;	// Shift sampling window
	  iBefore = 0;		// Reset index once shift has been performed

	  for (i = 0; i < nSamp[j]; ++i)	// For the records currently in our sample window
	    {
	      if ((inputTime[j][i] + t_offset) >= tEnd)
		{
		  iAfter = i;	// Index of the first data value after (or at) tEnd

		  if (debug > 2)
		    fprintf (stderr,
			     "inputTime[%d][%d] + t_offset >= tEnd; iAfter=%d; breaking\n",
			     j, i, i);

		  break;
		}
	    }

	  if (iBefore == iAfter && iAfter < (nSamp[j] - 1))
	    ++iAfter;

	  if (debug > 1)
	    {
	      printf ("Held Input Times: ");

	      for (i = 0; i < nSamp[j]; ++i)
		printf ("%11.2f ", (inputTime[j][i] + t_offset));

	      printf ("\n");

	      printf ("iBefore=%d, iAfter=%d\n", iBefore, iAfter);
	      fflush (stdout);
	    }

	  if ((inputTime[j][iAfter] + t_offset) < tEnd)	// If pointer for iAfter is not beyond tEnd,
	    break;		// then we do not have enough data to do the
	  // next interpolation, so we need to break 
	  // out of this function and wait for another
	  // read.


	  // Calculate interpolation weighting fraction for beginning of time window
	  frac =
	    (tStart -
	     (inputTime[j][iBefore] + t_offset)) / (inputTime[j][iBefore +
								 1] -
						    inputTime[j][iBefore]);

	  if (debug > 1)
	    {
	      printf ("frac(begin)=%f\n", frac);
	    }

	  if (frac < 0 || frac > 1.0)
	    {
	      printf ("Error: Invalid fraction ( tStart ).\n");
	      exit (0);
	    }

	  nOut = 0;		// Reset counter at beginning of time window


	  // Integration/averaging of feed latencies requires specialized calculation
	  // due to the bizarre practice of splitting some of the larger incoming 
	  // packets into multiple smaller sub-packets prior to transmitting them
	  // to the wave ring but only attaching the feed latency of the original
	  // packet to the first of the sub-packets (which means that the other 
	  // sub-packets are listed as having a feed latency of 0.0, which is not
	  // reflective of reality).


	  // Apply weighting and add interpolated values to arrays used for calculations
	  time[nOut] = tStart;
	  dlatency[nOut] =
	    frac * inputLat[j][iBefore + 1] + (1. -
					       frac) * inputLat[j][iBefore];
	  psize[nOut] =
	    frac * packet_size[j][iBefore + 1] + (1. -
						  frac) *
	    packet_size[j][iBefore];

	  if (feedLat[j][iBefore + 1] < flatmin
	      && feedLat[j][iBefore] >= flatmin)
	    {
	      flatency[nOut] = feedLat[j][iBefore];
	    }
	  else if (feedLat[j][iBefore + 1] >= flatmin
		   && feedLat[j][iBefore] < flatmin)
	    {
	      flatency[nOut] = feedLat[j][iBefore + 1];
	    }
	  else
	    {
	      flatency[nOut] =
		frac * feedLat[j][iBefore + 1] + (1. -
						  frac) * feedLat[j][iBefore];
	    }

	  ++nOut;


	  // Add points entirely within time window to the arrays used for calculations
	  for (i = iBefore + 1; i < iAfter; ++i)
	    {
	      time[nOut] = inputTime[j][i] + t_offset;
	      dlatency[nOut] = inputLat[j][i];

	      if (debug > 2)
		{
		  printf ("feedLat[%d][%d] = %f\n", j, i, feedLat[j][i]);
		}

	      flatency[nOut] = feedLat[j][i];
	      psize[nOut] = packet_size[j][i];
	      ++nOut;
	    }


	  // Re-calculate interpolation weighting fraction for end of time window
	  frac =
	    (tEnd -
	     (inputTime[j][iAfter - 1] + t_offset)) / (inputTime[j][iAfter] -
						       inputTime[j][iAfter -
								    1]);

	  //////////////////////////////////////////////////
	  time_since_last = tEnd - inputTime[j][iAfter - 1];
	  //////////////////////////////////////////////////

	  if (debug > 1)
	    {
	      printf ("frac(end)=%f\n", frac);
	    }

	  if (frac < 0 || frac > 1.0)
	    {
	      printf ("Error: Invalid fraction ( tEnd ).\n");
	      exit (0);
	    }


	  // Apply weighting and add interpolated values to arrays used for calculations
	  time[nOut] = tEnd;
	  dlatency[nOut] =
	    frac * inputLat[j][iAfter] + (1. - frac) * inputLat[j][iAfter -
								   1];

	  if (feedLat[j][iAfter + 1] < flatmin
	      && feedLat[j][iAfter] >= flatmin)
	    flatency[nOut] = feedLat[j][iAfter];
	  else if (feedLat[j][iAfter + 1] >= flatmin
		   && feedLat[j][iAfter] < flatmin)
	    flatency[nOut] = feedLat[j][iAfter + 1];
	  else
	    flatency[nOut] =
	      frac * feedLat[j][iAfter] + (1. - frac) * feedLat[j][iAfter -
								   1];

	  psize[nOut] =
	    frac * packet_size[j][iAfter] + (1. -
					     frac) * packet_size[j][iAfter -
								    1];
	  ++nOut;


	  if (debug > 1)
	    {
	      for (i = 0; i < nOut; ++i)
		{
		  printf ("time[%d]=%f, dlatency[%d]=%f, psize[%d]=%f \n",
			  i, time[i], i, dlatency[i], i, psize[i]);
		  // printf("time[%d]=%f, dlatency[%d]=%f, flatency[%d]=%f, psize[%d]=%f \n", 
		  //              i, time[i], i, dlatency[i], i, flatency[i], i, psize[i]);
		}
	    }


	  // Calculate mean latency as trapezoidal summation over current time window
	  outputDLatency = 0;
	  outputFLatency = 0;
	  avPacketSize = 0;

	  for (i = 1; i < nOut; ++i)
	    {
	      outputDLatency +=
		0.5 * (dlatency[i] + dlatency[i - 1]) * (time[i] -
							 time[i - 1]);
	      avPacketSize +=
		0.5 * (psize[i] + psize[i - 1]) * (time[i] - time[i - 1]);
	    }

	  outputDLatency /= interval;
	  avPacketSize /= interval;


	  /////////////////////////////////////////////////////////////////////////////////////////////////////////

	  // The first nonzero flatency is fref.  Find on first pass.
	  // On second pass, if flatency is 0, check to see if previous 
	  // flatency value is nonzero, and set to that.  Otherwise,
	  // set to fref.  Then proceed with the summation as normal.

	  double fref = 0.0;
	  int frefset = 0;

	  for (i = 0; i < nOut; ++i)
	    {
	      if (!frefset)
		{
		  if (flatency[i] > flatmin)
		    {
		      fref = flatency[i];
		      // printf("fref set: %s-%s fref=%f\n", tag[j].net, tag[j].sta, fref);
		      frefset = 1;
		    }
		}
	    }

	  if (flatency[0] < flatmin)
	    flatency[0] = fref;

	  for (i = 1; i < nOut; ++i)
	    {
	      if (flatency[i] < flatmin)
		{
		  if (flatency[i - 1] >= flatmin)
		    {
		      flatency[i] = flatency[i - 1];
		      // printf("flatency[%d]=flatency[%d]=%f\n", i, i-1, flatency[i]);
		    }
		  else
		    {
		      flatency[i] = fref;
		      // printf("flatency[%d]=%f (fref=%f)\n", i, flatency[i], fref);
		    }
		}
	    }

	  // printf("%s-%s fref=%f\n", tag[j].net, tag[j].sta, fref);

	  for (i = 1; i < nOut; ++i)
	    {
	      outputFLatency +=
		0.5 * (flatency[i] + flatency[i - 1]) * (time[i] -
							 time[i - 1]);
	    }
	  outputFLatency /= interval;

	  /////////////////////////////////////////////////////////////////////////////////////////////////////////



	  if (debug > 1)
	    {
	      printf
		("Tag=%d, iBefore=%d, iAfter=%d: refTime=%11.2f, outputDLatency= %7.4f, outputFLatency=%7.4f, nshift=%d\n",
		 j, iBefore, iAfter, refTime[j], outputDLatency,
		 outputFLatency, iBefore + 1);
	    }

	  if (debug)
	    {
	      printf ("Time: %11.2f, DLatency: %7.4f, FLatency: %7.4f\n",
		      refTime[j], outputDLatency, outputFLatency);
	      fflush (stdout);
	    }


	  // Convert output latency to integer rounded to the nearest millisecond
	  *long_data_out = (int) (1000 * outputDLatency + 0.5);


	  // Populate trout and apply naming convention rules
	  strcpy (trout->sta, tag[j].sta);
	  strcpy (trout->chan, tag[j].cha);
	  strcpy (trout->net, tag[j].net);
	  strcpy (trout->loc, tag[j].loc);
          if (LocationCode[0] != 0)
            {
	       strcpy (trout->loc, LocationCode);
            }

	  if (renameChannel)
	    {
	      trout->chan[0] = chan0;
	      if (trout->chan[1] == 'H')
		{
		  trout->chan[1] = chan1;
		}
	      if (trout->chan[1] == 'N')
		{
		  trout->chan[1] = chan1;
		}
	    }

	  trout->version[0] = v0;
	  trout->version[1] = v1;
	  trout->pinno = pin;
	  trout->nsamp = outNsamp;
	  trout->samprate = samprate;
	  trout->endtime = refTime[j];
	  trout->starttime = trout->endtime - increment;
	  strcpy (trout->datatype, datatype);
	  // strcpy(trout->quality, quality);
	  strcpy (trout->pad, pad);

	  outputLength = sizeof (TRACE2_HEADER) + 4 * trout->nsamp;


	  // Print output header
	  if (debug)
	    {
	      fprintf (stderr, "SNW Latency --> ");
	      fprintf (stderr, "%5s.%s.%s.%s (0x%x 0x%x) ",
		       trout->sta, trout->chan, trout->net, trout->loc,
		       trout->version[0], trout->version[1]);
	      fprintf (stderr, "%d %s %3d ", trout->pinno, trout->datatype,
		       trout->nsamp);
	      fprintf (stderr, "%6.4f", trout->samprate);
	      fprintf (stderr, " (%.4f) (%.4f) ", trout->starttime,
		       trout->endtime);
	      // fprintf( stderr, "%d", long_data_out[0]);
	      fprintf (stderr, " Host:[%s] ", hostName);
	      fprintf (stderr, " Packet Size:[%.1f] ", avPacketSize);
	      // fprintf( stderr, " Time Since Last:[] ");
	      fprintf (stderr, " Average Latencies:[DL:%.1fs, FL %.1fs] ",
		       outputDLatency, outputFLatency);

	      if (slept == 1)
		{
		  fprintf (stderr, " Slept for %f ", sleepTime);
		}

	      fprintf (stderr, "\n");
	      fflush (stderr);
	    }

	  if (slept == 1)
	    slept = 0;

	  trout->quality[0] = 0;
	  trout->quality[1] = 0;

	  if (sleptStatus[j] == 1)
	    trout->quality[1] = 1;


	  // printf("d_now=%f ", d_now);
	  // printf("scnl_lastPacketTime[%d]=%f\n", j, scnl_lastPacketTime[j]);
	  // printf("\noutputFLatency=%f\n", outputFLatency);

	  int parts = 0;
	  char outStr[4096];
	  char tempStr[1024];
	  strcpy (outStr, "");

	  if (avPacketSize && ((avPacketSize * 100.0) >= 5.0))
	    ++parts;
	  if (outputDLatency && ((outputDLatency * 100.0) >= 5.0))
	    ++parts;
	  if (time_since_last && time_since_last > 0.0)
	    ++parts;
	  if (outputFLatency && ((outputFLatency * 100.0) >= 5.0))
	    ++parts;

	  printf ("SNW STRING ==>  ");
	  sprintf (tempStr, "%s-%s:%d:", trout->net, trout->sta, parts);
	  strcat (outStr, tempStr);

	  if (avPacketSize && ((avPacketSize * 100.0) >= 5.0))
	    {
	      sprintf (tempStr, "\"%s - Packet Size\"=%.3f;", hostName,
		       avPacketSize);
	      strcat (outStr, tempStr);
	    }
	  if (outputFLatency && ((outputFLatency * 100.0) >= 5.0))
	    {
	      sprintf (tempStr, "\"%s - Secs Between Each SCNL Packet Group\"=%.3f;", hostName, outputFLatency);	// {{ Average FL }}
	      strcat (outStr, tempStr);
	    }
	  if (time_since_last && time_since_last > 0.0)
	    {
	      sprintf (tempStr,
		       "\"%s - Secs Since Last Packet Received\"=%.3f;",
		       hostName, time_since_last);
	      strcat (outStr, tempStr);
	    }
	  if (outputDLatency && ((outputDLatency * 100.0) >= 5.0))
	    {
	      sprintf (tempStr, "\"%s - Secs of Transport Latency\"=%.3f", hostName, outputDLatency);	// {{ Average DL }}
	      strcat (outStr, tempStr);
	    }
	  sprintf (tempStr, "\n");
	  strcat (outStr, tempStr);

	  printf ("%s", outStr);
	  fflush (stdout);

	  // client_send_cmd(sock, outStr);

	  /* 
	     sprintf(tempStr, " %s", trout->net);
	     strcat(outStr, tempStr);    
	     sprintf(tempStr, "-%-5s %2d %20f", trout->sta, parts, d_now);
	     strcat(outStr, tempStr);    
	     // sprintf(tempStr, " D:%8.3f", outputDLatency); // {{ Average DL }}
	     sprintf(tempStr, " %12.3f", outputDLatency); // {{ Average DL }}
	     strcat(outStr, tempStr);    
	     // sprintf(tempStr, " F:%8.3f ", outputFLatency); // {{ Average FL }}
	     sprintf(tempStr, " %12.3f ", outputFLatency); // {{ Average FL }}
	     strcat(outStr, tempStr);    
	     // sprintf(tempStr, " time_since_last:%8.3f", time_since_last);
	     sprintf(tempStr, " %12.3f", time_since_last);
	     strcat(outStr, tempStr);    
	     // sprintf(tempStr, " Size:%8.3f", avPacketSize);
	     sprintf(tempStr, " %12.3f", avPacketSize);
	     strcat(outStr, tempStr);    
	     sprintf(tempStr, "\n");
	     strcat(outStr, tempStr);    

	     printf("%s", outStr);
	     fflush(stdout);
	   */

	  int sockReturn;
	  sockReturn = client_send_cmd (sock, outStr);

	  if (sockReturn != OK)
	    {
	      fprintf (stderr, "Error: client_send_cmd(%d, %s) return = %d\n",
		       sock, outStr, sockReturn);
	      printf ("ERROR code on send: %d\n", smerr);
	      fflush (stdout);
	      exit (0);
	    }

	  // sockReturn = client_recv_cmd(sock, str);

	  // if(sockReturn != OK)
	  // {
	  //   printf("ERROR code on receive: %d\n", smerr);
	  //   fflush(stdout);
	  //   exit(0);
	  // }

	  // if(sockReturn == -1)
	  // {
	  //   fprintf(stderr, "Error: client_send_cmd(%d, %s) failed (return = %d)\n", sock, outStr, sockReturn);
	  // }

	  if (debug)
	    {
	      printf ("sendReturn = %d\n", sockReturn);
	      fflush (stdout);
	    }

	  /////////////////////////////////////////////////////////////////////////////////////////////////////////

	  sleptStatus[j] = 0;


	  // Increment the reference time
	  refTime[j] += increment;


	  // If we do not have data to fill in intermediate times, 
	  // shift the reference time that we are looking for.
	  if (refTime[j] < (inputTime[j][0] + t_offset))
	    {
	      refTime[j] = (int) ((inputTime[j][0] + t_offset) + increment);
	    }
	}			// end while (processing a time window)
    }				// end for (each tag)
  return 0;
}

/************************************************************************************/
